
#include <iostream.h>
#include <stdio.h>
#include "mpi.h"

extern void   srand48(long);
extern double drand48(void);

#define BORN 1
#define DIES 0

/* The Life function */
double life(int matrix_size, int ntimes, MPI::Intracomm& comm)
{
  int      rank, size ;
  int      next, prev ;
  int      i, j, k;
  int      mysize, sum ;
  int    **matrix, **temp, **addr ;
  double   totaltime, starttime, maxtime =0;

  /* Determine size and my rank in communicator */
  size = comm.Get_size();
  rank = comm.Get_rank();

  /* Set neighbors */
  if (rank == 0) 
    prev = MPI::PROC_NULL;
  else
    prev = rank-1;
  if (rank == size - 1)
    next = MPI::PROC_NULL;
  else
    next = rank+1;

  /* Determine my part of the matrix */
  mysize = matrix_size/size + ((rank < (matrix_size % size)) ? 1 : 0 ) ;

  /* allocate the memory dynamically for the matrix */
  matrix = (int **) new int[mysize+2];
  temp = (int **) new int[mysize+2];
  for (i = 0; i < mysize+2; i++) {
    matrix[i] = new int[matrix_size+2];
    temp[i] = new int[matrix_size+2];
  }

  /* Initialize the boundaries of the life matrix */
  for (j = 0; j < matrix_size+2; j++)
    matrix[0][j] = matrix[mysize+1][j] = 
      temp[0][j] = temp[mysize+1][j] = DIES ;

  for (i = 0; i < mysize+2; i++)
    matrix[i][0] = matrix[i][matrix_size+1] = 
      temp[i][0] = temp[i][matrix_size+1] = DIES ;

  /* Initialize the life matrix */
  for (i = 1; i <= mysize; i++)  {
//    srand48((long)(1000^(i-1+mysize))) ;
    for (j = 1; j<= matrix_size; j++)
//    if (drand48() > 0.5)  
        matrix[i][j] = BORN ;
//      else
//        matrix[i][j] = DIES ;
  }

  /* Play the game of life for given number of iterations */
  starttime = MPI::Wtime() ;
  for (k = 0; k < ntimes; k++) {
    MPI::Request      req[4];
    MPI::Status       status[4];

    //cout << rank << " :> send/recv\n";
    /* Send and receive boundary information */
    req[0]=comm.Isend(&matrix[1][0],matrix_size+2,MPI::INT,prev,0); 
    req[1]=comm.Irecv(&matrix[0][0],matrix_size+2,MPI::INT,prev,0);
    req[2]=comm.Isend(&matrix[mysize][0],matrix_size+2,MPI::INT,next,0);
    req[3]=comm.Irecv(&matrix[mysize+1][0],matrix_size+2,MPI::INT,next,0);
    
    MPI::Request::Waitall(4, req, status);
    //cout << rank << " :< send/recv\n";

    /* For each element of the matrix ... */ 
    for (i = 1; i <= mysize; i++) {
      for (j = 1; j < matrix_size+1; j++) {

        /* find out the value of the current cell */
        sum = matrix[i-1][j-1] + matrix[i-1][j] + matrix[i-1][j+1] 
          + matrix[i][j-1] + matrix[i][j+1] 
            + matrix[i+1][j-1] + matrix[i+1][j] + matrix[i+1][j+1] ;
        
        /* check if the cell dies or life is born */
        if (sum < 2 || sum > 3)
          temp[i][j] = DIES ;
        else if (sum == 3)
          temp[i][j] = BORN ;
        else
          temp[i][j] = matrix[i][j] ;
      }
    }
    //cout << rank << " : swap\n";
    
    /* Swap the matrices */
    addr = matrix ;
    matrix = temp ;
    temp = addr ;
  }
  //cout << rank << " : reduce" << endl;
  /* Return the average time taken/processor */
  totaltime = MPI::Wtime() - starttime;
  comm.Reduce (&totaltime, &maxtime, 1, MPI::DOUBLE, MPI::SUM, 0);

  //cout << rank << " : delete\n";
  /* free the memory dynamically for the matrix */
  for (i = 0; i < mysize+2; i++) {
    delete matrix[i];
    delete temp[i];
  }
  delete matrix; delete temp;


  return (maxtime/(double)size);
}

void thread_main(int argc, char *argv[])
{
  int rank, N, iters ;
  double time ;

  MPI::Init (argc, argv);
  rank = MPI::COMM_WORLD.Get_rank () ;

  /* If I'm process 0, determine the matrix size and # of iterations */
  if ( rank == 0 ) {
    printf("Matrix Size : ") ;
    scanf("%d",&N) ;
    printf("Iterations : ") ;
    scanf("%d",&iters) ;
  }

  /* Broadcast the size and # of iterations to all processes */
  MPI::COMM_WORLD.Bcast(&N, 1, MPI::INT, 0) ;
  MPI::COMM_WORLD.Bcast(&iters, 1, MPI::INT, 0);

  /* Call the life routine */
  time = life ( N, iters, MPI::COMM_WORLD );

  /* Print the total time taken */
  if (rank == 0)
    cout << rank << "  Life finished in "<< time/100 << "seconds" << endl;

  MPI::Finalize();
  //return (MPI::SUCCESS);
}

int main(int argc, char *argv[])
{
  MPI::Init (argc, argv, thread_main);
}
