#include <mpi.h>
#include <stdlib.h>
#include <iostream.h>

#define DEBUX(X) X

template<class EL>
class array2D
{
public: 
  typedef EL value_type;

public:
  array2D () : rows_(0), cols_(0), els_(0) {}
  array2D (int rows, int cols) { allocate(rows, cols); }
  ~array2D () {delete[] els_;}

  
  ///////////////////////////////////
  // info
  int rows() const { return rows_; }
  int cols() const { return cols_; }

  EL& index (int row, int col) { return els_[row*cols_ + col]; }
  const EL& index (int row, int col) const { return els_[row*cols_ + col]; }

  void allocate (int rows, int cols) 
    { rows_ = rows; cols_ = cols;  els_ = new EL[elements()];}

  EL* get_data() const { return els_; }
  int get_data_size () const { return elements() * sizeof (EL); }
  

  void operator = (const array2D<EL>& m)
    {
      for (int i=0; i < elements(); i++)
	els_[i] = m.els_[i];
    }
  void operator = (EL m)
    {
      for (int r=0; r < rows_; r++)
	for (int c=0; c < cols_; c++)
	  index(r,c) = m;
    }

  void block_multiply (const array2D<EL>& a, const array2D<EL>& b, int pass)
    {
      for (int r=0; r < a.rows_; r++)
	for (int c=0; c < b.cols_; c++)
	  for (int k=0; k < a.cols_; k++)
	    index(r + pass*cols_,c) += a.index(r,k) * b.index(k,c);
    }

  friend ostream& operator << (ostream&o, array2D<EL>&m)
  {
      for (int r=0; r < m.rows_; r++)
      {
	for (int c=0; c < m.cols_; c++)
	  o << ' ' << m.index(r,c) ;
        o << endl;
      }
      return o;
  }
  
private:
  long elements () { return (long)rows_*cols_; }

  EL* els_;
  int rows_;
  int cols_;
};



main(int argc, char **argv)
{
  int rank, size;
  double starttime, endtime;

  MPI_Init (&argc, &argv);

  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);

  int n = atoi (argv[1]);

  MPI_Bcast (&n, 1, MPI_INT, 0, MPI_COMM_WORLD);

  array2D<double> A (n / size, n);
  array2D<double> B (n, n / size);
  array2D<double> C (n, n / size);
  array2D<double> D (n / size, n);

  A = 0.1;
  B = 0.1;
  C = 0;

  if (rank == 0)
	starttime = MPI_Wtime();

  for (int p = 0; p < size; p++)
    {
      D = A;
      MPI_Bcast (D.get_data(), n * n / size, MPI_DOUBLE, p, MPI_COMM_WORLD);
      //DEBUX(if (rank == 0) cerr << rank << " : pass " << p << endl << D);
      C.block_multiply (D, B, p);
    }

  if (rank == 0)
  {
    endtime = MPI_Wtime();
    cout << "multiplied " << n << " by " << n << " in " 
	<< endtime-starttime << endl;
  }

  MPI_Finalize();  
}



