#include <stdio.h>
#include <stdlib.h>

#include "mpi.h"


#define HELLO_TAG 100

int hello_alltoall (MPI::Comm& comm)
{
  int rank,size, dest;
  int  msg_length;
  char msg_buffer[64];
  char rcv_buffer[64];
  MPI::Status status;

  /* Determine the size of this group and the current rank */
  rank = comm.Get_rank();
  size = comm.Get_size();

  sprintf (msg_buffer, "hello from %d", rank);
  msg_length = strlen (msg_buffer)+1;
  
  /* Send the message to the other nodes */
  int round = 0;
  for (dest = (rank+1)%size; dest != rank; dest = (dest + 1)%size, round++)
    {
      if (rank%2) 
	{
	  comm.Send (msg_buffer, msg_length, MPI::CHAR,
		     dest, HELLO_TAG);
	  comm.Recv (rcv_buffer, 64,  MPI::CHAR, 
		     MPI::ANY_SOURCE, HELLO_TAG, status);
	}
      else
	{
	  comm.Recv (rcv_buffer, 64,  MPI::CHAR, 
		     MPI::ANY_SOURCE, HELLO_TAG, status);
	  comm.Send (msg_buffer, msg_length, MPI::CHAR,
		     dest, HELLO_TAG);
	}
      cout << rank << ": round " << round << " : got " << rcv_buffer << endl;
    }
  return 0;
}
		  
		  
void
thread_main(int argc, char**argv)
{
  MPI::Init (argc, argv);
  int count = 10;
  double starttime, endtime ;

  if (MPI::COMM_WORLD.Get_rank () == 0)
    starttime = MPI::Wtime();
 
  for (int i = 0; i < count; i++)
  {
     hello_alltoall (MPI::COMM_WORLD);
  }
  
  if (MPI::COMM_WORLD.Get_rank () == 0)
    {
      endtime = MPI::Wtime();
      cerr << count << " alltoall in " << endtime - starttime << endl;
    }
  

  MPI::Finalize();

}
int main(int argc, char**argv)
{
  MPI::Init (argc, argv, thread_main);
}
  


  
