// -*- c++ -*- ------------------------------------------
// Title: Comm
// Author: Kristian Kvilekval
// Project:


#include "mpi.h"

MPI::Request MPI::Comm::Isend(const void *buf, int count,
		     const Datatype & datatype, int dest, int tag) const
{
  int rank = Get_rank();
  Request rq ((void*)buf, count , datatype.get_size(), tag, 
	      rank, dest, &group);


  if (dest == PROC_NULL || count == 0)
    {
      rq.set_status (MPI::Envelope::DONE);
      return rq;
    }

  //DEBUGX(cerr << rank << " : posting send" << endl);
  rq.post_send ();
  return rq;
}


MPI::Request MPI::Comm::Irecv(void *buf, int count,
		    const Datatype & datatype, int source, int tag) const
{
  int rank = Get_rank();
  Request rq (buf, count , datatype.get_size(), tag, 
	      source , rank, &group);

  if (source == PROC_NULL || count == 0)
    {
      rq.set_status (MPI::Envelope::DONE);
      return rq;
    }
  
  //DEBUGX(cerr << rank << " : posting rcv" << endl);
  rq.post_recv ();
  return rq;
}

void MPI::Comm::Send(const void *buf, int count, 
		const Datatype & datatype, int dest, int tag) const
{
//   DEBUGX(cerr << Get_rank () << " : Send -> " << dest << endl);
  Request rq = Isend (buf, count , datatype, dest, tag);
  rq.Wait();
}

void MPI::Comm::Recv(void *buf, int count, const Datatype & datatype,
		int source, int tag, Status & status) const
{
  Request rq  = Irecv (buf, count, datatype, source, tag);
  rq.Wait(status);
}

void MPI::Comm::Recv(void *buf, int count, const Datatype & datatype,
		int source, int tag) const
{
//   DEBUGX(cerr << Get_rank () << " : Recv <- " << source << endl);
  Request rq  = Irecv(buf, count , datatype, source , tag);
  rq.Wait();
}



bool MPI::Comm::Iprobe(int source, int tag, Status & status) const
{
  // Find a request for me from SOURCE with tag TAG
  return false;
}
bool MPI::Comm::Iprobe(int source, int tag) const
{
  return false;
}
void MPI::Comm::Probe(int source, int tag, Status & status) const
{
}
void MPI::Comm::Probe(int source, int tag) const
{
}
