// -*- c++ -*- ------------------------------------------
// Title: Comm_queue
// Author: Kristian Kvilekval
// Project:

#include <algo.h>

#include "mpi.h"


MPI::Comm_queue::Comm_queue() 
{
  //PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
  pthread_mutex_init (&gate, NULL);
  pthread_cond_init (&changed, NULL);
  changes = 0;
}
MPI::Comm_queue::~Comm_queue() 
{
  //  DEBUGX(cerr << "~Comm_queue" << endl);
  assert (queue_.size() == 0);
    
}

void MPI::Comm_queue::add_envelope(Envelope&e)
{
//     DEBUGX(cerr << "adding  " << e << endl);
  queue_.insert (queue_.end(), e);
  //  task->signal_change();
}

void MPI::Comm_queue::rem_envelope(envelope e)
{
  queue_.erase (e);
}

bool MPI::Comm_queue::send_rendezvous (Envelope&snd)
{
  Envelope rcv;
  snd.set_status (MPI::Envelope::SEND_READY);

//   DEBUGX(cerr << "send    " << snd << endl);

  lock ();
  snd.set_mutex (&gate);
  snd.set_cond (&changed);
  if (find_matching_receiver (snd, rcv))
    {
      rendezvous (snd, rcv);
      unlock();
      return true;
    }

  add_envelope (snd);
  unlock();
  return false;
}

bool MPI::Comm_queue::recv_rendezvous (Envelope&dst)
{
  Envelope snd;

  dst.set_status (MPI::Envelope::RECV_READY);

//   DEBUGX(cerr << "recv    " << dst << endl);

  lock();

  dst.set_mutex (&gate);
  dst.set_cond (&changed);

  if (find_matching_sender (dst, snd))
    {
      rendezvous (snd, dst);
      unlock();
      return true;
    }
  add_envelope (dst);

  unlock();
  return false; 
}

bool MPI::Comm_queue::rendezvous (Envelope&src, Envelope&dst)
{
  memcpy (dst.get_buffer(), src.get_buffer(), src.get_length());

  src.set_status (MPI::Envelope::DONE);
  dst.set_status (MPI::Envelope::DONE);
  src.signal_tasks();
  dst.signal_tasks();

//     DEBUGX ( cerr << "rendez " << dst <<"<-" << src << endl);

  return true;
}

struct recv_match {
  recv_match (const MPI::Envelope &m) : sender_(m) {}
  bool operator () (MPI::Envelope &t) { 
    return ((t.get_status() == MPI::Envelope::RECV_READY)
// 	    && (t.get_destination() == sender_.get_destination())//Always true
	    && (t.get_source () == MPI::ANY_SOURCE
		||t.get_source() == sender_.get_source())
	    && (t.get_tag() == MPI::ANY_TAG
		|| t.get_tag() == sender_.get_tag()));
  }
  const MPI::Envelope &sender_;
};


MPI::Comm_queue::Queue::iterator 
recvfind_if (MPI::Comm_queue::Queue::iterator first, 
	     MPI::Comm_queue::Queue::iterator last, 
	     recv_match pred)
{				
    while (first != last && !pred(*first)) ++first;
    return first;
}

bool MPI::Comm_queue::find_matching_receiver (const Envelope &snd,
					      Envelope &rcv )
{
  envelope  e =  recvfind_if (queue_.begin(), queue_.end(), 
			      recv_match (snd));
  
  if (!null_envelope (e)) 
    {
      rcv = *e;
      rem_envelope(e);
      return true;
    }
  
  return false;
}

struct send_match 
{
  send_match (const MPI::Envelope &m) : recvr_(m) {}
  bool operator () (MPI::Envelope& t) { 
    return ((t.get_status() == MPI::Envelope::SEND_READY)
// 	    && (t.get_destination() == m_.get_destination())//Always true
	    && (recvr_.get_source () == MPI::ANY_SOURCE
		|| t.get_source() == recvr_.get_source())
	    && (recvr_.get_tag () == MPI::ANY_TAG
		|| t.get_tag() == recvr_.get_tag()));
  }
  const MPI::Envelope& recvr_;
};


MPI::Comm_queue::Queue::iterator 
sendfind_if (MPI::Comm_queue::Queue::iterator first, 
	     MPI::Comm_queue::Queue::iterator last, 
	     send_match pred)
{				
    while (first != last && !pred(*first)) ++first;
    return first;
}

bool MPI::Comm_queue::find_matching_sender (const Envelope &rcv,
					    Envelope &snd )
{
   envelope  e = sendfind_if (queue_.begin(), queue_.end(), 
			      send_match (rcv));

  
    if (!null_envelope (e)) 
      {
	snd = *e;
	rem_envelope (e);
	return true;
      }

    return false;
}



