// -*- c++ -*- ------------------------------------------------
// mpi
// 
// Author : Kristian Kvilekval
// Project:
// Date   : May 1 1998

#include <sys/time.h>
#include <string.h>
#include "mpi.h"

const MPI::Datatype MPI::BYTE (MPI::Datatype::DT_BYTE, sizeof (char));
const MPI::Datatype MPI::CHAR (MPI::Datatype::DT_CHAR, sizeof (char));
const MPI::Datatype MPI::SHORT (MPI::Datatype::DT_SHORT, sizeof (short));

const MPI::Datatype MPI::BOOLEAN (MPI::Datatype::DT_BOOLEAN, sizeof (bool));
const MPI::Datatype MPI::INT (MPI::Datatype::DT_INT, sizeof (int));
const MPI::Datatype MPI::LONG (MPI::Datatype::DT_LONG, sizeof (long));
const MPI::Datatype MPI::FLOAT (MPI::Datatype::DT_FLOAT, sizeof (float));
const MPI::Datatype_double MPI::DOUBLE;

const MPI::Datatype MPI::TWOINT (MPI::Datatype::DT_INT, 2*sizeof (int));

MPI::Intracomm MPI::COMM_WORLD(0);

const MPI::Op_max MPI::MAX (MPI::Op::OP_MAX);
const MPI::Op_min MPI::MIN (MPI::Op::OP_MIN);
const MPI::Op_sum MPI::SUM (MPI::Op::OP_SUM);

thread_control MPI::TCB;


void MPI::Init (int& argc, char**& argv)
{
  COMM_WORLD.Barrier();
}

void MPI::Init (int& argc, char**& argv, MPI::mpi_void_main mf)
{
  // Scan comamnd line for number of threads wanted.
  // Generate tasks.

  MPI::Init (argc, argv, (mpi_int_main)mf);
}

void MPI::Init (int& argc, char**& argv, MPI::mpi_int_main mf)
{
  // Construct initial objects.
  int num_tasks = 4;

  for (int i=1; i < argc; i++)
    {
      if (strcmp (argv[i], "-np") == 0) 
	{
	  num_tasks = atoi (argv[++i]);
	  // Shift the rest down and cover these argument up
	  memmove (&argv[i-1], &argv[i+1], (argc-i)*sizeof (char*));
	  argc-=2;
	}
    }

  TCB.init(num_tasks);
  
  COMM_WORLD.set_barrier_size(num_tasks);
  
  Group::initialize_group (((Intracomm&)COMM_WORLD).get_group(),
			   num_tasks,
			   mf, argc, argv);

//   DEBUGX (cerr << "main_thread : wait_empty" << endl);
  TCB.wait_empty();
//   DEBUGX (cerr << "main_thread : exiting" << endl);
}

void MPI::Finalize()
{
//   DEBUGX (cerr << Get_current_task()->get_rank () << ": Finalize" << endl);
  COMM_WORLD.Barrier();
}

MPI::Task* MPI::Get_current_task ()
{
  return MPI::Task::get_current();
}

double MPI::Wtime ()
{
  struct timeval tv;
  gettimeofday (&tv, 0);
  return (tv.tv_sec*1e6 + tv.tv_usec) / 1e6;
}
