// -*- c++ -*- ------------------------------------------------
// mpi.h
// 
// Author : Kristian Kvilekval
// Date   : May 1 1998
#ifndef __MPI_H
#define __MPI_H


#include <pthread.h>
#include <list.h>
#include <function.h>
#include <vector.h>
#include "tcb.h"

#define MPI_NAMESPACE 0

#if MPI_NAMESPACE
namespace MPI {
#define MPI_EXTERN_DATA extern
#define MPI_FUNCTION 
#else
#define MPI_EXTERN_DATA static
#define MPI_FUNCTION static
class MPI {
public:
#endif

class Comm;
class Intracomm;		// public Comm
class Graphcomm;		// public Intracomm
class Cartcomm;			// public Intracomm
class Intercomm;		// public Comm
class Datatype;
class Errhandler;
class Exception;
class Group;
class Op;
class Request;
class Prequest;			// public Request
class Status;
//
class Envelope;
class Comm_man;
class Comm_queue;
class Task;

// SendRecv constants
enum { ANY_SOURCE = -32768,
       ANY_ADDR  , 
       ANY_TAG  , 
       PROC_NULL   
};

// Compare operations values
enum { IDENT = 1,
       EQUAL = 2,
       SIMILAR = 3,
       UNEQUAL = 4
};

// typedef void User_function(const void* invec, void* inoutvec, int len,
// 			   const Datatype& datatype);

typedef int (*mpi_int_main) (int,char**);
typedef void (*mpi_void_main) (int,char**);

typedef int Aint;

#include "exception.h"
#include "op.h"
#include "datatype.h"
#include "request.h"
#include "status.h"
#include "group.h"
#include "comm.h"
#include "intracomm.h"

#include "envelope.h"
#include "comm_queue.h"
#include "mpi_task.h"

// collective operations
MPI_EXTERN_DATA const Op_max MAX;
MPI_EXTERN_DATA const Op_min MIN;
MPI_EXTERN_DATA const Op_sum SUM;
// MPI_EXTERN_DATA const Op PROD;
// MPI_EXTERN_DATA const Op MAXLOC;
// MPI_EXTERN_DATA const Op MINLOC;
// MPI_EXTERN_DATA const Op BAND;
// MPI_EXTERN_DATA const Op BOR;
// MPI_EXTERN_DATA const Op BXOR;
// MPI_EXTERN_DATA const Op LAND;
// MPI_EXTERN_DATA const Op LOR;
// MPI_EXTERN_DATA const Op LXOR;

// Default datatypes.
MPI_EXTERN_DATA const Datatype BYTE;
MPI_EXTERN_DATA const Datatype CHAR;
MPI_EXTERN_DATA const Datatype SHORT;
MPI_EXTERN_DATA const Datatype BOOLEAN;
MPI_EXTERN_DATA const Datatype INT;
MPI_EXTERN_DATA const Datatype LONG;
MPI_EXTERN_DATA const Datatype FLOAT;
MPI_EXTERN_DATA const Datatype_double DOUBLE;

MPI_EXTERN_DATA const Datatype TWOINT;

MPI_EXTERN_DATA Intracomm COMM_WORLD;


#if 0
// null handles
MPI_EXTERN_DATA const Group        GROUP_NULL;
//MPI_EXTERN_DATA const Comm         COMM_NULL;
//MPI_EXTERN_DATA const MPI_Comm     COMM_NULL;
MPI_EXTERN_DATA Comm_Null          COMM_NULL;
MPI_EXTERN_DATA const Datatype     DATATYPE_NULL;
MPI_EXTERN_DATA Request            REQUEST_NULL;
MPI_EXTERN_DATA const Op           OP_NULL;
MPI_EXTERN_DATA const Errhandler   ERRHANDLER_NULL;  
#endif


MPI_FUNCTION  void Init (int& argc, char** &argv);
MPI_FUNCTION  void Init (int& argc, char** &argv, mpi_void_main);
MPI_FUNCTION  void Init (int& argc, char** &argv, mpi_int_main);
MPI_FUNCTION  void Finalize();
MPI_FUNCTION double Wtime();

MPI_FUNCTION  MPI::Task *Get_current_task();

#if 0
MPI_FUNCTION_ void Attach_buffer(void* buffer, int size);
MPI_FUNCTION_ int Detach_buffer(void*& buffer);

MPI_FUNCTION_ void Compute_dims(int nnodes, int ndims, int dims[]);
MPI_FUNCTION_ void Get_processor_name(char*& name, int& resultlen);
MPI_FUNCTION_ void Get_error_string(int errorcode, char*& string, int& resultlen);
MPI_FUNCTION_ int Get_error_class(int errorcode);

MPI_FUNCTION_ double Wtick();
MPI_FUNCTION_ bool Is_initialized();
#endif



MPI_EXTERN_DATA thread_control TCB;

#if MPI_NAMESPACE
}
#else
private:
 MPI() {}
}; 
#endif

#define DEBUGX(X) X
//#define DEBUGX(X)

#endif       // __MPI_H
