// -*- c++ -*- ------------------------------------------
// Title: MPI_task
// Author: Kristian Kvilekval
// Project:

#include "mpi.h"


pthread_key_t MPI::Task::task_key;

void MPI::Task::init_task_key() 
{
  pthread_key_create (&task_key, 0/*delete fun*/);
}


MPI::Task::Task (mpi_int_main f, int ac, char**av) 
  : main_thread(f), argc(ac), argv(av),
    recv_queue ()
{
  static pthread_once_t task_key_once = PTHREAD_ONCE_INIT;

  pthread_once (&task_key_once, Task::init_task_key);

  pthread_mutex_init (&gate, 0);
  pthread_cond_init (&change, 0);
  changes = 0;
}

MPI::Task::~Task ()
{
}

void MPI::Task::action (int r)
{
  rank = r;
  pthread_setspecific(task_key, (void*)this);

  main_thread (argc, argv);
}

MPI::Task* MPI::Task::get_current()
{
  return (Task*)pthread_getspecific (task_key);
}

void MPI::Task::lock()
{
  pthread_mutex_lock (&gate);
}
void MPI::Task::unlock()
{
  pthread_mutex_unlock (&gate);
}

void MPI::Task::wait_change()
{
  while (changes == 0)
    pthread_cond_wait (&change, &gate);

  changes = 0;
}
void MPI::Task::signal_change()
{
  changes++;
  pthread_cond_broadcast (&change);
}
