// -*- c++ -*- ----------------------------------------------------
//  tcb.h
//   Kristian Kvilekval
//  CS240b
//  thread control block
//  main thread header for threaded application
// ---------------------------------------------------------------------
#ifndef _TCB_H
#define _TCB_H

#include <pthread.h>
#include <stdarg.h>
#include <vector.h>
#include <deque.h>
#include <list.h>
#include <stack.h>

//#ifdef __sgi
#define SGISTL
//#endif

class thread_control;
class thread_task;
class thread_worker;

/////////////////////////////////////////////////////////////
// Base class for user defined tasks.  User
// needs to overide ACTION method.
class thread_task
{
public:
  thread_task () {}
  virtual ~thread_task ();
  virtual void action(int rank);
};

////////////////////////////////////////////////////////////
// Allow pointers in an STL container
struct task_wrap
{
  thread_task *ptr;
  task_wrap () : ptr(0) {}
  task_wrap (thread_task*p) :ptr(p) {}
};

////////////////////////////////////////////////////////////
// Thread routine
class thread_worker
{
public:
  thread_worker() : TCB(0),rank(0){}
  thread_worker(thread_control*t, int r) : TCB(t), rank(r) {}
  
  void worker();

private:
  thread_control *TCB;
  int rank;
};

extern void *worker_thread(void*);

////////////////////////////////////////////////////
// Worker queue manangement.
class thread_control
{
public:  
  thread_control ();
  //~thread_control();

  void init (int threads);
  int Threads() { return total_threads; }

  void push_task(thread_task *t);
  //thread_task pop_task();

  void wait_empty();

  void broadcast (void *, int sz, int sender);

private:
  friend class thread_task;
  friend class thread_worker;
  
  pthread_mutex_t start_lock;
  pthread_cond_t start_cond;
  pthread_mutex_t stop_lock;
  pthread_cond_t stop_cond;

#ifdef SGISTL
  queue<thread_worker, list <thread_worker> > worker_queue;
  queue<task_wrap, list <task_wrap> > task_queue;
#else
  queue<list <thread_worker> > worker_queue;
  queue<list <task_wrap> > task_queue;
#endif
  
  int  threads_running;
  int  total_threads;
  int  working;
  int  partition;
};

////////////////////////////////////////////////////////
// Barrier object.
// NOTE:  There must be only one for all threads using
// barrier
class thread_barrier
{
public:
  thread_barrier (int tot, char *name=0) 
    : total(tot), count(0), leaving(0), named(name?name:"") 
    {
      init_barrier();
    }

  void wait();

  void set_size (int x) { total = x; }
  int  get_size () { return total; }
private:
  int init;
  void init_barrier();

  pthread_mutex_t count_lock;
  pthread_mutex_t gate_lock;
  pthread_cond_t  gate_cond;
  pthread_cond_t  leaving_cond;

  int total;
  int count;
  int leaving;
  char *named;
};

/////////////////////////////////////////////////////////////
// Thread broadcast.
// NOTE:  There must be only one for all threads using
// boradcast object.

class thread_broadcast
{
public:
  thread_broadcast (int threads, char *name=0) 
    : bar(threads, name) {}

  void broadcast (void *, int size, int sender, int rank);

private:
  thread_barrier bar;
  void * sender_data;
};
 

#endif










