#include <iostream.h>
#include "mpi.h"


void user_bcast(int buffer[], int count, MPI::Intracomm comm);
int thread_main(int argc, char **argv);

int 
main(int argc, char **argv)
{
  MPI::Init (argc, argv, thread_main);
}

int 
thread_main(int argc, char **argv)
{
  int rank, size;
  int msg[10];

  // Start up MPI 
  MPI::Init(argc, argv);

  rank = MPI::COMM_WORLD.Get_rank();
  size = MPI::COMM_WORLD.Get_size();
 
  // Create an array to broadcast on the last process 
  if (rank == size - 1) {
    cout << "Broadcast: ";
    for (int i=0; i<10; i++) {
      msg[i] = i;
      cout << " " << msg[i] << " ";
    }
    cout << endl;
  }

  double starttime = MPI::Wtime();
  // Do the broadcast 
  user_bcast(msg, 10, MPI::COMM_WORLD);

  double endtime  = MPI::Wtime();
  // If we are the console, print what we got 
  if (rank == 0) {
    cout << "process  " << rank << " got:";
    for (int i=0; i<10; i++)
      cout << " " << msg[i] << " ";
    cout << endl;
    cout << " in " << endtime - starttime << endl;
  }

  // Quit MPI 
  MPI::Finalize();

  return 0;
}


void
user_bcast(int buffer[], int count, MPI::Intracomm comm)
{
  int rank = comm.Get_rank();
  int size = comm.Get_size();
  
  /* The last process loops to send to everyone in the communicator. */
  /* Do a Waitall to wait for all messages to finish sending */

  if (rank == size - 1 ) {

    MPI::Request* reqarray = new MPI::Request[size-1];
    
    for (int dest = 0; dest < size-1; dest++ )
      reqarray[dest] = comm.Isend(buffer,count, MPI::INT, dest, 4);

    MPI::Request::Waitall(size-1, reqarray);

  } 

  /* If we are not the last process process, we should go into a receive */
  /* and wait for it to complete */

  else {
    MPI::Request request;
    
    request = comm.Irecv(buffer, count, MPI::INT, size - 1, MPI::ANY_TAG);
    request.Wait();
  }
}







