// -*- c++ -*- ------------------------------------------
// Title: Comm
// Author: Kristian Kvilekval
// Project:


#include "mpi.h"



MPI::Comm::~Comm()
{
  //  DEBUGX(cerr << "~comm" << endl);
}


MPI::Request 
MPI::Comm::Isend(const void *buf, int count,
		 const Datatype & datatype, 
		 int dest, int tag)const
{
  int source = Get_rank ();

  if (dest == PROC_NULL || count == 0 || source == dest)
    {
      dest = source;
    }

  MPI::Task* local = group->index(source);
  MPI::Task* remote = group->index (dest);
  // Sends are posted to the remote queue
  Comm_queue::envelope ev = 
    remote->send_invitation((void*)buf, count * datatype.get_size(), 
			    datatype.get_type(), tag, source, dest, local);

  Request rq (ev);

  if (source == dest)
    {
      rq.set_completed(MPI::Envelope::DONE);
    }

  //DEBUGX(cerr << rank << " : posting send" << endl);
  return rq;
}


MPI::Request 
MPI::Comm::Irecv(void *buf, int count, const Datatype & datatype, 
		 int source, int tag) const
{
  int dest = Get_rank ();
  MPI::Task* local = group->index(dest);
      
  // Recieves are posted to the local queue
  Comm_queue::envelope ev = 
    local->recv_invitation(buf, count * datatype.get_size(), 
			   datatype.get_type(), tag, source, dest, local);
  Request rq (ev);

  if (source == PROC_NULL || count == 0 || source == dest)
    {
      rq.set_completed(MPI::Envelope::DONE);
    }

  return rq;
}

void MPI::Comm::Send(const void *buf, int count, 
		const Datatype & datatype, int dest, int tag) const
{
//   DEBUGX(cerr << Get_rank () << " : Send -> " << dest << endl);
  Request rq = Isend (buf, count , datatype, dest, tag);
  rq.Wait();
}

void MPI::Comm::Recv(void *buf, int count, const Datatype & datatype,
		int source, int tag, Status & status) const
{
  Request rq  = Irecv (buf, count, datatype, source, tag);
  rq.Wait(status);
}

void MPI::Comm::Recv(void *buf, int count, const Datatype & datatype,
		int source, int tag) const
{
//   DEBUGX(cerr << Get_rank () << " : Recv <- " << source << endl);
  Request rq  = Irecv(buf, count , datatype, source , tag);
  rq.Wait();
}



bool MPI::Comm::Iprobe(int source, int tag, Status & status) const
{
  // Find a request for me from SOURCE with tag TAG
  return false;
}
bool MPI::Comm::Iprobe(int source, int tag) const
{
  return false;
}
void MPI::Comm::Probe(int source, int tag, Status & status) const
{
}
void MPI::Comm::Probe(int source, int tag) const
{
}
