// -*- c++ -*- ------------------------------------------
// Title: Comm
// Author: Kristian Kvilekval
// Project:
#ifndef _Comm_H
#define _Comm_H

class Comm_manager;

class Comm {
public:
  Comm() { group = 0; }
  Comm(const Comm& x) : group (x.group) {}
  //Comm(const Group&x) : group (x) {}
  virtual ~Comm();
  //////////////////////////////////////////////////
  // Groups, Contexts, and Communicators  
  virtual Group& Get_group() const { return *group; }
  virtual int   Get_size() const { return group->Get_size(); }
  virtual int   Get_rank() const { return group->Get_rank(); }
  virtual void  Free() { delete group; }
#if 0
  static int    Compare(const Comm & comm1, const Comm & comm2);
  virtual Comm& Clone() const = 0;
  virtual bool  Is_inter() const;
  virtual int   Get_topology() const;
#endif

  //////////////////////////////////////////////////
  // point to point
  virtual Request Isend(const void *buf, int count,
			const Datatype & datatype, int dest, int tag) const;
  virtual Request Irecv(void *buf, int count,
			const Datatype & datatype, int source, int tag) const;

  virtual void Send(const void *buf, int count, 
		    const Datatype & datatype, int dest, int tag) const;
  virtual void Recv(void *buf, int count, const Datatype & datatype,
		    int source, int tag, Status & status) const;
  virtual void Recv(void *buf, int count, const Datatype & datatype,
		    int source, int tag) const;

  virtual bool Iprobe(int source, int tag, Status & status) const;
  virtual bool Iprobe(int source, int tag) const;

  virtual void Probe(int source, int tag, Status & status) const;
  virtual void Probe(int source, int tag) const;
#if 0  
  virtual void Sendrecv(const void *sendbuf, int sendcount,
			const Datatype & sendtype, int dest, int sendtag, 
			void *recvbuf, int recvcount, 
			const Datatype & recvtype, int source,
			int recvtag, Status & status) const;
  
  virtual void Sendrecv(const void *sendbuf, int sendcount,
			const Datatype & sendtype, int dest, int sendtag, 
			void *recvbuf, int recvcount, 
			const Datatype & recvtype, int source,
			int recvtag) const;

  virtual void Sendrecv_replace(void *buf, int count,
				const Datatype & datatype, int dest, 
				int sendtag, int source,
				int recvtag, Status & status) const;

  virtual void Sendrecv_replace(void *buf, int count,
				const Datatype & datatype, int dest, 
				int sendtag, int source,
				int recvtag) const;
#endif

#if 0
  ////////////////////////////////////////////////
  // Buffered
  virtual void Bsend(const void *buf, int count,
		     const Datatype & datatype, int dest, int tag) const;
  virtual void Ssend(const void *buf, int count, 
		     const Datatype & datatype, int dest, int tag) const ;
  virtual void Rsend(const void *buf, int count,
		     const Datatype & datatype, int dest, int tag) const;
  virtual Request Ibsend(const void *buf, int count, const
			 Datatype & datatype, int dest, int tag) const;
  virtual Request Issend(const void *buf, int count,
			 const Datatype & datatype, int dest, int tag) const;
  virtual Request Irsend(const void *buf, int count,
			 const Datatype & datatype, int dest, int tag) const;
#endif
protected:
  
  friend class MPI;
  void set_group (Group *g) { group = g; }
  Group* get_group () { return group; }
  
  Group *group;

private:
};

#endif  // _Comm_H
