// -*- c++ -*- ------------------------------------------
// Title: Comm_queue
// Author: Kristian Kvilekval
// Project:

#include <algo.h>

#include "mpi.h"

DEBUGX(int MPI::Comm_queue::debug = 0;)

MPI::Comm_queue::Comm_queue() 
{
  //PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP
  pthread_mutex_init (&gate, NULL);
  pthread_cond_init (&changed, NULL);
  changes = 0;
}

MPI::Comm_queue::~Comm_queue() 
{
  //DEBUGX(cerr << "~Comm_queue" << endl);
  assert (queue_.size() == 0);
    
}

// void MPI::Comm_queue::add_envelope(Envelope&e)
// {
// //     DEBUGX(cerr << "adding  " << e << endl);
//   //queue_.insert (queue_.end(), e);
//   //  task->signal_change();
// }

// void MPI::Comm_queue::rem_envelope(envelope e)
// {
//   queue_.erase (e);
// }

bool MPI::Comm_queue::send_rendezvous (Envelope&snd)
{
  snd.set_status (MPI::Envelope::SEND_READY);

   DEBUGT(Comm_queue, cerr << snd.get_task()->get_rank() 
	 << " : send : " << snd << endl);

  envelope rcv = find_matching_receiver (snd);

  if (!null_envelope (rcv))
    {
      rendezvous (snd, *rcv);

      snd.set_status (MPI::Envelope::DONE);
      snd.task_signal_complete();

      (*rcv).task_lock();
      (*rcv).set_status (MPI::Envelope::DONE);
      (*rcv).task_signal_complete();
      (*rcv).task_unlock();

      return true;
    }
  //add_envelope (snd);

  return false;
}

bool MPI::Comm_queue::recv_rendezvous (Envelope&dst)
{

  dst.set_status (MPI::Envelope::RECV_READY);

   DEBUGT(Comm_queue, cerr << dst.get_task()->get_rank() 
	  << " : recv : " << dst << endl);

  envelope snd = find_matching_sender (dst);
  if (!null_envelope(snd))
    {
      rendezvous (*snd, dst);

      dst.set_status (MPI::Envelope::DONE);
      dst.task_signal_complete();

      (*snd).task_lock();
      (*snd).set_status (MPI::Envelope::DONE);
      (*snd).task_signal_complete();
      (*snd).task_unlock();

      return true;
    }
  //add_envelope (dst);

  return false; 
}

bool MPI::Comm_queue::rendezvous (Envelope&src, Envelope&dst)
{
  memcpy (dst.get_buffer(), src.get_buffer(), src.get_length());

  DEBUGT(Comm_queue, cerr <<  "  : rdvz : " << dst <<"<-" << src << endl);

  return true;
}

struct recv_match {
  recv_match (const MPI::Envelope &m) : sender_(m) {}
  bool operator () (MPI::Envelope &t) { 
    return ((t.get_status() == MPI::Envelope::RECV_READY)
// 	    && (t.get_destination() == sender_.get_destination())//Always true
	    && (t.get_source () == MPI::ANY_SOURCE
		||t.get_source() == sender_.get_source())
	    && (t.get_tag() == MPI::ANY_TAG
		|| t.get_tag() == sender_.get_tag()));
  }
  const MPI::Envelope &sender_;
};

/////////////////////////////////////////////////////
// This is HACK!
// SGI's stl would not compile find_if... so I reprouduce
// it here
MPI::Comm_queue::Queue::iterator 
recvfind_if (MPI::Comm_queue::Queue::iterator first, 
	     MPI::Comm_queue::Queue::iterator last, 
	     recv_match pred)
{				
    while (first != last && !pred(*first)) ++first;
    return first;
}

MPI::Comm_queue::envelope
MPI::Comm_queue::find_matching_receiver (const Envelope &snd)
{
  return  recvfind_if (queue_.begin(), queue_.end(), 
		       recv_match (snd));
}

/////////////////////////////////////////////////
// A sender match code.
struct send_match 
{
  send_match (const MPI::Envelope &m) : recvr_(m) {}
  bool operator () (MPI::Envelope& t) { 
    return ((t.get_status() == MPI::Envelope::SEND_READY)
// 	    && (t.get_destination() == m_.get_destination())//Always true
	    && (recvr_.get_source () == MPI::ANY_SOURCE
		|| t.get_source() == recvr_.get_source())
	    && (recvr_.get_tag () == MPI::ANY_TAG
		|| t.get_tag() == recvr_.get_tag()));
  }
  const MPI::Envelope& recvr_;
};


/////////////////////////////////////////////////////
// This is HACK!
// SGI's stl would not compile find_if... so I reprouduce
// it here
MPI::Comm_queue::Queue::iterator 
sendfind_if (MPI::Comm_queue::Queue::iterator first, 
	     MPI::Comm_queue::Queue::iterator last, 
	     send_match pred)
{				
    while (first != last && !pred(*first)) ++first;
    return first;
}

MPI::Comm_queue::envelope
MPI::Comm_queue::find_matching_sender (const Envelope &rcv)
{
  return  sendfind_if (queue_.begin(), queue_.end(), 
		       send_match (rcv));
}



