// -*- c++ -*- ------------------------------------------
// Title: Envelope
// Author: Kristian Kvilekval
// Project:
#ifndef _Envelope_H
#define _Envelope_H

class Envelope 
{
public:
  enum envelope_status { CANCELLED=-1, 
			 INIT = 0, SEND_READY = 1, 
			 RECV_READY = 2, DONE = 3 };

public:

  Envelope() {}
  //  Envelope(const Envelope&e) { handle = e.handle; handle->inc_ref(); }
  //  ~Envelope() { if (handle->dec_ref() == 0) delete handle; }

  Envelope(void*data, int len, int type, int tg, int src, int dst)
    : buffer (data), length (len), typeTag(type), tag(tg), 
      source(src), destination (dst),
       context (0), status (0)
    {
    }

  /////////////////////////////////////////////
  // Access
  int get_destination() const { return destination;} 
  int get_source() const { return source;} 
  void* get_context() const { return context;} 
  int get_length() const { return length;} 
  int get_status() const { return status;} 
  int get_tag() const { return tag;} 
  int get_typeTag() const { return typeTag;}
  void *get_buffer() const {  return buffer; }
  Task* get_task () { return task; }

  void set_destination(int x) { destination = x; };
  void set_source(int x) { source = x; };
  void set_context(void* x) { context = x; };
  void set_length(int x) { length = x; }; 
  void set_status(int x) { status = x; };
  void set_tag(int x) { tag = x; };
  void set_typeTag(int x) { typeTag = x; };
  void set_buffer(void *x) {  buffer = x; }
  void set_task (Task*x) { task = x; }

  void task_lock();
  void task_unlock();
  void task_signal_complete ();

  friend ostream& operator << (ostream&o, const Envelope&);

private:
  void *buffer;
  int length;           // Length of data to send/recv.
  int typeTag;          // Datatype.typeTag (BYTE...)
  int tag;              // MPI user TAG.
  int source;
  int destination;
  void* context;          // Communicator context.
  int status;           // Used by protocol
  Task*task;
};

//#include <pthread.h>
#if 0
class Envelope_impl
{
private:
  friend class Envelope;

public:
  Envelope_impl (int x) : ref_count (x) {}
  Envelope_impl (void*data, int len, int type, int tg, int src, int dst)
    : ref_count (0), buffer (data), length (len), typeTag(type), tag(tg), 
      source(src), destination (dst),
       context (0), status (0)
    {
    }

  int inc_ref () { return ++ref_count; }
  int dec_ref () { return (--ref_count >= 0)?ref_count:0; }
private:
  int ref_count;

  void *buffer;
  int length;           // Length of data to send/recv.
  int typeTag;          // Datatype.typeTag (BYTE...)
  int tag;              // MPI user TAG.
  int source;
  int destination;
  void* context;          // Communicator context.
  int status;           // Used by protocol

  pthread_mutex_t *mutex;
  pthread_cond_t *condition;
};

MPI_EXTERN_DATA Envelope_impl gNullEnvelope;

class Envelope 
{
public:
  enum envelope_status { CANCELLED=-1, 
			 INIT = 0, SEND_READY = 1, 
			 RECV_READY = 2, DONE = 3 };

public:

  Envelope() { handle = &gNullEnvelope; handle->inc_ref (); }
  Envelope(const Envelope&e) { handle = e.handle; handle->inc_ref(); }
  ~Envelope() { if (handle->dec_ref() == 0) delete handle; }

  Envelope(void*data, int len, int type, int tag, int src, int dst)
    {
      handle = new Envelope_impl (data, len, type,  tag, src, dst);
      handle->inc_ref();
    }

  Envelope& operator = (const Envelope &e) { 
    if (handle->dec_ref() == 0) delete handle;
    handle = e.handle;
    handle->inc_ref();
    return *this;
  }

  /////////////////////////////////////////////
  // Access
  int get_destination() const { return handle->destination;} 
  int get_source() const { return handle->source;} 
  void* get_context() const { return handle->context;} 
  int get_length() const { return handle->length;} 
  int get_status() const { return handle->status;} 
  int get_tag() const { return handle->tag;} 
  int get_typeTag() const { return handle->typeTag;}
  void *get_buffer() const {  return handle->buffer; }

  void set_destination(int x) { handle->destination = x; };
  void set_source(int x) { handle->source = x; };
  void set_context(void* x) { handle->context = x; };
  void set_length(int x) { handle->length = x; }; 
  void set_status(int x) { handle->status = x; };
  void set_tag(int x) { handle->tag = x; };
  void set_typeTag(int x) { handle->typeTag = x; };
  void set_buffer(void *x) {  handle->buffer = x; }

  //void set_waiting (pthread_cond_t *c) { handle->waiting = c; }

  void set_mutex (pthread_mutex_t*x) { handle->mutex = x; }
  void set_cond (pthread_cond_t*x) { handle->condition = x; }
  void lock();
  void unlock();
  void signal_tasks ();
  void wait ();

  friend ostream& operator << (ostream&o, const Envelope&e)
    {
      o << "envelope(" 
	<< "src=" << e.get_source() << ','
	<< "dst=" << e.get_destination() << ','
	<< "tag=" << e.get_tag() << ','
	<< "stat=" << e.get_status() 
	<< ')';
      return o;
    }

private:

  Envelope_impl *handle;
};
#endif



#endif  // _Envelope_H
