// -*- c++ -*- --------------------------------------------------
// intracomm.h
//

#include "mpi.h"

#define TAG_REDUCE (-1)


MPI::Intracomm::~Intracomm ()
{
  //DEBUGX(cerr << "~Intracomm" << endl);
}


void MPI::Intracomm::Barrier() 
{
//   DEBUGX (cerr << "Barrier : " << barrier.get_size() << endl);
  barrier.wait ();
}

void MPI::Intracomm::Bcast(void *buffer, int count, 
			   const Datatype& datatype, int root)
{
  int rank = Get_rank();
  int size = Get_size();

  // Do checks
  //DEBUGX (cerr << rank << " : enter Bcast "  << endl);

  if (root == PROC_NULL) return;
  if (size == 0 || size == 1) return;

  if (rank == root)
    bcast_data = buffer;

  Barrier();

  //DEBUGX (cerr << rank  << ":  Bcast copy : "  << endl);
  
  if (rank != root) memcpy (buffer, bcast_data, datatype.Get_size()*count);

  Barrier();
  //DEBUGX (cerr << rank  << ": exit Bcast : "  << endl);
}


void MPI::Intracomm::Reduce(const void *sendbuf, void *recvbuf, int count, 
			    const Datatype & datatype, const Op & op, 
			    int root)
{
  int rank = Get_rank();
  int size = Get_size();

  //DEBUGX (cerr << rank << ": enter reduce : " << size << endl);
  // Do checks

  if (root == PROC_NULL) return;
  if (size == 0 ) return;

  if (rank != root)
    {
      Send (sendbuf, count, datatype, root, TAG_REDUCE);
      return;
    }


  // Initialize buffer
  if (rank == (size - 1))
    {
      //DEBUGX (cerr << rank << ": intracomm "   " copy" << endl);
      memcpy (recvbuf, sendbuf, datatype.Get_size() * count);
    }
  else
    {
      //DEBUGX (cerr << rank<< ": intracomm "   " recv" << endl);
      Recv (recvbuf, count, datatype, size - 1, TAG_REDUCE);
    }

  // Allocate a recv buffer;
  void *inbuf = new char [datatype.Get_size () * count];
  for (int i = size - 2; i >= 0 ; --i)
    {
      if (i == rank)
	{
 	  //DEBUGX (cerr << rank<< ": intracomm op" << count << endl);
	  op.operation (sendbuf, recvbuf, datatype, count);
	}
      else
	{
 	  //DEBUGX (cerr << rank<< ": intracomm recv-op" << count << endl);
	  Recv (inbuf, count, datatype, i, TAG_REDUCE);

	  op.operation (inbuf, recvbuf, datatype, count);
	}
    }
  delete[] inbuf;
}
  
