// -*- c++ -*- ------------------------------------------------
// mpi
// 
// Author : Kristian Kvilekval
// Project:
// Date   : May 1 1998

#include <sys/time.h>
#include <string.h>
#include "mpi.h"

#undef DEBUGT
#define DEBUGT(X) if(MPI::debug) do { X; } while(0)


DEBUG_DECL(int MPI::debug = 0;)

const MPI::Datatype MPI::BYTE (MPI::Datatype::DT_BYTE, sizeof (char));
const MPI::Datatype MPI::CHAR (MPI::Datatype::DT_CHAR, sizeof (char));
const MPI::Datatype MPI::SHORT (MPI::Datatype::DT_SHORT, sizeof (short));

const MPI::Datatype MPI::BOOLEAN (MPI::Datatype::DT_BOOLEAN, sizeof (bool));
const MPI::Datatype MPI::INT (MPI::Datatype::DT_INT, sizeof (int));
const MPI::Datatype MPI::LONG (MPI::Datatype::DT_LONG, sizeof (long));
const MPI::Datatype MPI::FLOAT (MPI::Datatype::DT_FLOAT, sizeof (float));
const MPI::Datatype_double MPI::DOUBLE;

const MPI::Datatype MPI::TWOINT (MPI::Datatype::DT_INT, 2*sizeof (int));

MPI::Intracomm MPI::COMM_WORLD;

const MPI::Op_max MPI::MAX (MPI::Op::OP_MAX);
const MPI::Op_min MPI::MIN (MPI::Op::OP_MIN);
const MPI::Op_sum MPI::SUM (MPI::Op::OP_SUM);



void MPI::Init (int& argc, char**& argv)
{
  COMM_WORLD.Barrier();
}

void MPI::Init (int& argc, char**& argv, MPI::mpi_void_main mf)
{
  // Scan comamnd line for number of threads wanted.
  // Generate tasks.

  MPI::Init (argc, argv, (mpi_int_main)mf);
}


int MPI::process_options (int& argc, char**& argv)
{
  int num_tasks = 4;
  for (int i=1; i < argc;)
    {
      if (strcmp (argv[i], "-np") == 0) 
	{
	  num_tasks = atoi (argv[i+1]);
	  // Shift the rest down and cover these argument up
	  memmove (&argv[i], &argv[i+2], (argc-i-1)*sizeof (char*));
	  argc-=2;
	}
      else if (strcmp (argv[i], "-drequest")==0)
	{
	  MPI::Request::debug = 1;
	  // Shift the rest down and cover these argument up
	  memmove (&argv[i], &argv[i+1], (argc-i)*sizeof (char*));
	  argc-=1;
	}
      else if (strcmp (argv[i], "-dcomm_queue")==0)
	{
	  MPI::Comm_queue::debug = 1;
	  // Shift the rest down and cover these argument up
	  memmove (&argv[i], &argv[i+1], (argc-i)*sizeof (char*));
	  argc-=1;
	}
      else if (strcmp (argv[i], "-dmpi")==0)
	{
	  MPI::debug = 1;
	  // Shift the rest down and cover these argument up
	  memmove (&argv[i], &argv[i+1], (argc-i)*sizeof (char*));
	  argc-=1;
	}
      else
	i++;
    }
  return num_tasks;
}
void MPI::Init (int& argc, char**& argv, MPI::mpi_int_main mf)
{
  // Construct initial objects.
  int num_tasks = process_options (argc, argv);

  COMM_WORLD.set_barrier_size(num_tasks);

  thread_control *TCB = new thread_control;

  TCB->init(num_tasks);
  
  Group *group = new Group;

  //Group::initialize_group (*group, num, mf, argc, argv);

  for (int i = 0; i < num_tasks; i++)
    {
      MPI::Task *tsk = new MPI::Task (mf, argc, argv);
      tsk->set_rank (i);
      group->tasks.insert (group->tasks.end(), Group::task_ptr (tsk));
      TCB->push_task (tsk);
    }
  COMM_WORLD.set_group(group);

  TCB->start_work();

//   DEBUGX (cerr << "main_thread : wait_empty" << endl);
  TCB->wait_empty();
//   DEBUGX (cerr << "main_thread : exiting" << endl);
}

void MPI::Finalize()
{
  DEBUGT (cerr << Get_current_task()->get_rank () << " : ->Finalize" << endl);
  COMM_WORLD.Barrier();
  DEBUGT (cerr << Get_current_task()->get_rank () << " : <-Finalize" << endl);
}

MPI::Task* MPI::Get_current_task ()
{
  return MPI::Task::get_current();
}

double MPI::Wtime ()
{
  struct timeval tv;
  gettimeofday (&tv, 0);
  return (tv.tv_sec*1e6 + tv.tv_usec) / 1e6;
}
