// -*- c++ -*- ------------------------------------------
// Title: MPI_task
// Author: Kristian Kvilekval
// Project:

#include "mpi.h"


pthread_key_t MPI::Task::task_key;

void MPI::Task::init_task_key() 
{
  pthread_key_create (&task_key, 0/*delete fun*/);
}


MPI::Task::Task (mpi_int_main f, int ac, char**av) 
  : main_thread(f), argc(ac), argv(av),
    recv_queue ()
{
  static pthread_once_t task_key_once = PTHREAD_ONCE_INIT;

  pthread_once (&task_key_once, Task::init_task_key);

  pthread_mutex_init (&requests.gate, 0);
  pthread_cond_init (&requests.completed, 0);
  requests.completed_count = 0;
}

MPI::Task::~Task ()
{
  //cerr << "~Task" << endl;
}

void MPI::Task::action (int r)
{
  pthread_setspecific(task_key, (void*)this);

  main_thread (argc, argv);
}

MPI::Task* MPI::Task::get_current()
{
  return (Task*)pthread_getspecific (task_key);
}



MPI::Comm_queue::envelope
MPI::Task::recv_invitation(void*data, int len, int type, 
			   int tag, int src, int dst, 
			   Task*notify)
{
  recv_queue.lock();
  Comm_queue::envelope e = 
    recv_queue.alloc_envelope(data,len,type,tag,src,dst);
  (*e).set_task (notify);
  recv_queue.recv_rendezvous (*e);
  recv_queue.unlock();
  return e;
}

MPI::Comm_queue::envelope
MPI::Task::send_invitation(void*data, int len, int type, 
			   int tag, int src, int dst,
			   Task*notify)
{
  recv_queue.lock();
  Comm_queue::envelope e =
    recv_queue.alloc_envelope(data,len,type,tag,src,dst);
  (*e).set_task (notify);
  recv_queue.send_rendezvous (*e);
  recv_queue.unlock();
  return e;
}    

void 
MPI::Task::kill_invitation(Comm_queue::envelope x)
{
  recv_queue . lock();
  recv_queue . rem_envelope (x);
  recv_queue . unlock();
}

