// -*- c++ -*- ------------------------------------------
// Title: MPI_task
// Author: Kristian Kvilekval
// Project:
#ifndef _MPI_task_H
#define _MPI_task_H

class Task : public thread_task
{
public:
  Task(mpi_int_main f, int ac, char**av) ;
  ~Task();

  /////////////////////////////////////////////
  // Thread attributes;
  int get_rank () { return rank; }
  void set_rank(int x) { rank = x; }

  static Task* get_current();	// Current executing thread;
  void action (int rank);	// MPI main (overrides thread_task);

  /////////////////////////////////////////////
  // Recieve queue
  Comm_queue& get_recv_queue () { return recv_queue; }
  void        del_envelope(Comm_queue::envelope x);

  /////////////////////////////////////////
  // Request handling
  void lock()   {pthread_mutex_lock (&requests.gate);}
  void unlock() {pthread_mutex_unlock (&requests.gate);}  
  void wait_completed() {pthread_cond_wait (&requests.completed, &requests.gate);}
  void signal_completed(){pthread_cond_broadcast (&requests.completed);}
  int  completed() { return requests.completed_count; }
  int  inc_completed() { return ++requests.completed_count; }
  int  dec_completed() { return --requests.completed_count; }

  ////////////////////////////////////////
  // Invitations
  Comm_queue::envelope  recv_invitation(void*data, int len, int type, 
					int tag, int src, int dst, 
					Task*notify);

  Comm_queue::envelope  send_invitation(void*data, int len, int type, 
					int tag, int src, int dst,
					Task*notify);

  void kill_invitation(Comm_queue::envelope x);
  
private:
  static void init_task_key() ;
  static pthread_key_t task_key;

  mpi_int_main main_thread;
  int     argc;
  char**  argv;
  Comm_queue recv_queue;
  int rank;

  struct completions {
    pthread_mutex_t gate;
    pthread_cond_t  completed;
    int     completed_count;
  } requests;
};	// end of class task

#endif  // _MPI_task_H
