// -*- c++ -*----------------------------------------------
// Title: request
// Project:
// Author:
// Date:

#include "mpi.h"


DEBUG_DECL(int MPI::Request::debug = 0;)
MPI::Request::Request ()
{}
MPI::Request::~Request()
{}


void MPI::Request::Wait(Status &status)
{
  assert (envelope != 0);
  MPI::Task * task = MPI::Get_current_task ();


  DEBUGT(Request, cerr << task->get_rank() << " : Wait : enter " << *envelope << endl);
  task->lock();
  for (;;)
    {
       DEBUGT (Request, cerr << task->get_rank() 
	       << " : Wait : completed " << task->completed() << endl);
      while (task->completed() == 0)
	task->wait_completed();

      if (get_status() == MPI::Envelope::DONE)
	{
	  task->dec_completed();
	  task->unlock();
 	  DEBUGT(Request, cerr << task->get_rank() << " : Wait : leave" << endl);
	  status.set_from_envelope(get_envelope());
	  reclaim_envelope();
	  return;
	}
      task->wait_completed();
    }
}

void MPI::Request::Wait()
{
  Status s;
  Wait (s);
}

bool MPI::Request::Test(Status &status)
{
  return false;
}

bool MPI::Request::Test()
{
  return false;
}

void MPI::Request::Free(void)
{
}

void MPI::Request::Cancel(void) const
{
}

int MPI::Request::Waitany(int count, MPI::Request reqs[], Status& status)
{
  MPI::Task * task = MPI::Get_current_task ();

  int index = -1;
  //bool found = false;

  task->lock();
  for (;;) 
    {
      while (task->completed() == 0)
	task->wait_completed();
      
      for (index = 0; index < count; ++index)
	{
	  if (reqs[index].get_status() == MPI::Envelope::DONE)
	    {
	      task->dec_completed();
	      task->unlock();
	      status.set_from_envelope(reqs[index].get_envelope());
	      reqs[index].reclaim_envelope();
	      return index;
	    }
	}
      task->wait_completed();
    }
}

int MPI::Request::Waitany(int count, MPI::Request reqs[])
{
  Status status;
  return Waitany (count, reqs, status);
}

void MPI::Request::Waitall(int count, MPI::Request reqs[], Status stats[])
{
  for (int i = 0; i < count; ++i)
    reqs[i].Wait (stats[i]);
}
void MPI::Request::Waitall(int count, MPI::Request reqs[])
{
  for (int i = 0; i < count; ++i)
    reqs[i].Wait ();
}

#if 0
bool MPI::Request::Testany(int count, MPI::Request reqs[], int& index, Status& status);
bool MPI::Request::Testany(int count, MPI::Request reqs[], int& index);
bool MPI::Request::Testall(int count, MPI::Request reqs[], Status stats[]);
bool MPI::Request::Testall(int count, MPI::Request reqs[]);
int MPI::Request::Waitsome(int count, MPI::Request reqs[],
		      int array_of_indices[], Status stats[]) ;
int MPI::Request::Waitsome(int count, MPI::Request reqs[],
		      int array_of_indices[]);
int MPI::Request::Testsome(int count, MPI::Request reqs[],
		      int array_of_indices[], Status stats[]);
int MPI::Request::Testsome(int count, MPI::Request reqs[],
		      int array_of_indices[]);
#endif

//////////////////////////////////////////////////////////////////
#if 0
void MPI::Request::post_send()
{
  // Find the task queue of the recv'er
  MPI::Group *g = (MPI::Group*)envelope.get_context();
  MPI::Task  *t = g->index (envelope.get_destination());
  t->get_recv_queue().send_rendezvous (envelope);
}

void MPI::Request::post_recv()
{
  // Find the task queue of the recv'er
  MPI::Group *g = (MPI::Group*)envelope.get_context();
  MPI::Task  *t = g->index (envelope.get_destination());
  t->get_recv_queue().recv_rendezvous (envelope);
}
#endif
